﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace ClosedLoopTestCSharp
{
    public partial class Form1 : Form
    {
        #region Fields

        AcqCallBackDelegate acqCb;
        Stopwatch beginSw = new Stopwatch();
        long firstDataIndex = -1;
        long lastDataIndex = -1;
        double maxIntervalMs;
        double intervalLimitMs = 10;
        int countMissed;
        int countIntervalAboveLimit;
        double previousCallbackTimeStamp;

        #endregion

        #region Constructors
        public Form1()
        {
            InitializeComponent();
        }
        #endregion

        #region Event Handlers

        private void SubscribeButton_Click(object sender, EventArgs e)
        {
            acqCb = new AcqCallBackDelegate(AcqCallBack);
            IntPtr funcPtr = Marshal.GetFunctionPointerForDelegate(acqCb);
            if (NativeCalls.Subscribe(funcPtr))
            {
                ResetStats();
                subscribeButton.Enabled = false;
                unsubscribeButton.Enabled = true;
                readInfoButton.Enabled = true;
            }
        }

        private void unSubscribeButton_Click(object sender, EventArgs e)
        {
            NativeCalls.UnSubscribe();
            Thread.Sleep(250);
            firstDataIndex = -1;
            lastDataIndex = -1;
            maxIntervalMs = 0;
            subscribeButton.Enabled = true;
            unsubscribeButton.Enabled = false;
            readInfoButton.Enabled = false;
            ClearInfo();
        }

        private void clearButton_Click(object sender, EventArgs e)
        {
            ResetStats();
        }

        private void readInfoButton_Click(object sender, EventArgs e)
        {
            ClearInfo();

            nBuffersTextBox.Text = NativeCalls.GetNBuffers().ToString();

            samplingRateTextBox.Text = NativeCalls.GetSamplingRate().ToString("#.00");
            mvOffsetTextBox.Text = NativeCalls.GetMVOffset().ToString("#.00");
            adcCountsToMvTextBox.Text = NativeCalls.GetADCCountsToMV().ToString("#.00");

            int nrows = NativeCalls.GetArrayNRows();
            nRowsTextBox.Text = nrows.ToString();
            int ncols = NativeCalls.GetArrayNCols();
            nColsTextBox.Text = ncols.ToString();

            int nAcqChs = NativeCalls.GetNAcqChannels();
            nChsTextBox.Text = nAcqChs.ToString();

            IntPtr pCh = NativeCalls.GetIndexesOfAcqChannels();
            int[] acqChIndexes = new int[nAcqChs];
            Marshal.Copy(pCh, acqChIndexes, 0, nAcqChs);
            for (int i = 0; i < acqChIndexes.Length; i++)
                chIndexesListBox.Items.Add(acqChIndexes[i]);

            //int test1 = NativeCalls.GetAcqFramePositionOfChannelIndex(acqChIndexes[64]);
            //int test2 = NativeCalls.GetAcqFramePositionOfChannelSubscripts(2, 1);

            rowComboBox.Items.Clear();
            for (int i = 0; i < nrows; i++)
                rowComboBox.Items.Add(i);

            colComboBox.Items.Clear();
            for (int i = 0; i < ncols; i++)
                colComboBox.Items.Add(i);
        }

        private void rowColComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (rowComboBox.SelectedItem != null && colComboBox.SelectedItem != null)
            {
                int row = (int)rowComboBox.SelectedItem;
                int col = (int)colComboBox.SelectedItem;

                acqIndexTextBox.Text = NativeCalls.GetAcqFramePositionOfChannelSubscripts(row, col).ToString();
            }
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (unsubscribeButton.Enabled)
                NativeCalls.UnSubscribe();
        }

        #endregion

        #region Methods
        void AcqCallBack(long index, IntPtr data, int dataLength)
        {
            if (firstDataIndex == -1)
            {
                firstDataIndex = lastDataIndex = index;
                previousCallbackTimeStamp = 0;
                beginSw.Restart();
            }

            if (index > lastDataIndex + 1)
                this.BeginInvoke((MethodInvoker)(() => missedDataTextBox.Text = (++countMissed).ToString()));
            lastDataIndex = index;

            byte[] dataArray = new byte[dataLength];
            Marshal.Copy(data, dataArray, 0, dataLength);

            DoWork(dataArray);

            if (lastDataIndex > firstDataIndex)
            {
                double currentCallbackTimestamp = beginSw.Elapsed.TotalMilliseconds;
                double elapsed = currentCallbackTimestamp - previousCallbackTimeStamp;
                previousCallbackTimeStamp = currentCallbackTimestamp;

                long nCallbacks = lastDataIndex - firstDataIndex;
                double averageInterval = beginSw.Elapsed.TotalMilliseconds / nCallbacks;

                this.BeginInvoke((MethodInvoker)(() => 
                    {
                        if (intervalsListBox.Items.Count <= 1000)
                            intervalsListBox.Items.Add(elapsed.ToString("0.00") + " ms");

                        if (nCallbacks % 100 == 0)
                            averageIntervalTextBox.Text = averageInterval.ToString("0.00") + " ms";

                        if (elapsed > maxIntervalMs)
                        {
                            maxIntervalMs = elapsed;
                            maxIntervalTextBox.Text = maxIntervalMs.ToString("0.00") + " ms";
                        }
                        if (elapsed > intervalLimitMs)
                        {
                            numberAboveLimitTextBox.Text = (++countIntervalAboveLimit).ToString();
                        }
                    }));
            }
        }

        private void DoWork(byte[] dataArray)
        {
            int msPause = 0;
            this.Invoke((MethodInvoker)(() => msPause = (int)pauseNumericUpDown.Value));
            Thread.Sleep(msPause);
        }

        void ResetStats()
        {
            firstDataIndex = -1;
            maxIntervalMs = 0;
            countMissed = 0;
            countIntervalAboveLimit = 0;
            maxIntervalTextBox.Text = " - ";
            numberAboveLimitTextBox.Text = "0";
            missedDataTextBox.Text = "0";
            intervalsListBox.Items.Clear();
        }
        void ClearInfo()
        {
            nBuffersTextBox.Text = "";

            mvOffsetTextBox.Text = "";
            adcCountsToMvTextBox.Text = "";

            nChsTextBox.Text = "";

            chIndexesListBox.Items.Clear();
        }
        #endregion

        #region Classes and Delegates

        delegate void AcqCallBackDelegate(long index, IntPtr data, int dataLength);

        class NativeCalls
        {
            [DllImport("BWClosedLoop.dll")]
            internal static extern bool Subscribe(IntPtr cb);
            [DllImport("BWClosedLoop.dll")]
            internal static extern bool UnSubscribe();
            [DllImport("BWClosedLoop.dll")]
            internal static extern  int GetNBuffers();
            [DllImport("BWClosedLoop.dll")]
            internal static extern double GetBufferLevel();
            [DllImport("BWClosedLoop.dll")]
            internal static extern double GetSamplingRate();
            [DllImport("BWClosedLoop.dll")]
            internal static extern double GetMVOffset();
            [DllImport("BWClosedLoop.dll")]
            internal static extern double GetADCCountsToMV();
            [DllImport("BWClosedLoop.dll")]
            internal static extern int GetArrayNRows();
            [DllImport("BWClosedLoop.dll")]
            internal static extern int GetArrayNCols();
            [DllImport("BWClosedLoop.dll")]
            internal static extern int GetNAcqChannels();
            [DllImport("BWClosedLoop.dll")]
            internal static extern IntPtr GetIndexesOfAcqChannels();
            [DllImport("BWClosedLoop.dll")]
            internal static extern int ChannelSubscriptsToIndex(int row, int col);
            [DllImport("BWClosedLoop.dll")]
            internal static extern void ChannelIndexToSubscripts(int index, ref int row, ref int col);
            [DllImport("BWClosedLoop.dll")]
            internal static extern int GetAcqFramePositionOfChannelIndex(int index);
            [DllImport("BWClosedLoop.dll")]
            internal static extern int GetAcqFramePositionOfChannelSubscripts(int row, int col);
        }

        #endregion
    }
}
